class PluginError(Exception):
    """An exception raised for plugin-related errors."""


class PluginValidationError(PluginError):
    """An exception raised when a plugin package is not valid."""


class InvalidPluginFormat(PluginValidationError):
    """An exception raised when the plugin file format is not supported."""


class PluginInstallationError(PluginError):
    """An exception raised when a plugin fails to install."""



from datetime import datetime
from enum import Enum
from typing import Any, cast

from canvas_sdk.effects.base import EffectType, _BaseEffect


class TaskStatus(Enum):
    """TaskStatus."""

    COMPLETED = "COMPLETED"
    CLOSED = "CLOSED"
    OPEN = "OPEN"


class AddTask(_BaseEffect):
    """
    An Effect that will create a Task in Canvas.
    """

    class Meta:
        effect_type = EffectType.CREATE_TASK
        apply_required_fields = ("title",)

    assignee_id: str | None = None
    patient_id: str | None = None
    title: str | None = None
    due: datetime | None = None
    status: TaskStatus = TaskStatus.OPEN
    labels: list[str] = []

    @property
    def values(self) -> dict[str, Any]:
        """The values for Task addition."""
        return {
            "patient": {"id": self.patient_id},
            "due": self.due.isoformat() if self.due else None,
            "assignee": {"id": self.assignee_id},
            "title": self.title,
            "status": self.status.value,
            "labels": self.labels,
        }


class AddTaskComment(_BaseEffect):
    """
    An Effect that will create a Task Comment on a Task.
    """

    class Meta:
        effect_type = EffectType.CREATE_TASK_COMMENT
        apply_required_fields = (
            "body",
            "task_id",
        )

    body: str | None = None
    task_id: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The values for adding a task comment."""
        return {"task": {"id": self.task_id}, "body": self.body}


class UpdateTask(_BaseEffect):
    """
    An Effect that will update a Task in Canvas.
    """

    class Meta:
        effect_type = EffectType.UPDATE_TASK
        apply_required_fields = ("id",)

    id: str | None = None
    assignee_id: str | None = None
    patient_id: str | None = None
    title: str | None = None
    due: datetime | None = None
    status: TaskStatus = TaskStatus.OPEN
    labels: list[str] = []

    @property
    def values(self) -> dict[str, Any]:
        """The values for adding a task comment."""
        value_dict: dict[str, Any] = {}
        # Only add the fields that have been explicitly set on the model (exclude_unset=True).
        # Oherwise, the effect interpreter will set values to null based on their defaults.
        set_fields = self.model_dump(exclude_unset=True)
        for field, val in set_fields.items():
            if field.endswith("_id"):
                value_dict[field.split("_")[0]] = {"id": val}
            elif field == "due" and val is not None:
                value_dict[field] = cast(datetime, val).isoformat()
            elif field == "status":
                value_dict[field] = cast(TaskStatus, val).value
            else:
                value_dict[field] = getattr(self, field)
        return value_dict



from canvas_sdk.effects import Effect, EffectType
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol


class Protocol(BaseProtocol):
    """
    You should put a helpful description of this protocol's behavior here.
    """

    # Name the event type you wish to run in response to
    RESPONDS_TO = EventType.Name(EventType.UNKNOWN)

    NARRATIVE_STRING = "I was inserted from my plugin's protocol."

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        return [Effect(type=EffectType.LOG, payload="Hello, world!")]



from collections import defaultdict
from typing import Union, cast

from django.utils.functional import classproperty


class CodeConstants:
    """A class representing different code systems and their URLs."""

    CPT = "CPT"
    HCPCS = "HCPCS"
    HCPCSLEVELII = "HCPCSLEVELII"
    CVX = "CVX"
    LOINC = "LOINC"
    SNOMEDCT = "SNOMEDCT"
    FDB = "FDB"
    RXNORM = "RXNORM"
    ICD10CM = "ICD10CM"
    ICD10PCS = "ICD10PCS"
    NUCC = "NUCC"
    CANVAS = "CANVAS"
    INTERNAL = "INTERNAL"
    NDC = "NDC"

    URL_CPT = "http://www.ama-assn.org/go/cpt"
    URL_HCPCS = "http://www.cms.gov/medicare/coding/medhcpcsgeninfo"
    URL_HCPCSLEVELII = "https://coder.aapc.com/hcpcs-codes"
    URL_CVX = "http://hl7.org/fhir/sid/cvx"
    URL_LOINC = "http://loinc.org"
    URL_SNOMEDCT = "http://snomed.info/sct"
    URL_FDB = "http://www.fdbhealth.com/"
    URL_RXNORM = "http://www.nlm.nih.gov/research/umls/rxnorm"
    URL_ICD10 = "ICD-10"
    URL_NUCC = "http://www.nucc.org/"
    URL_CANVAS = "CANVAS"
    URL_INTERNAL = "INTERNAL"
    URL_NDC = "http://hl7.org/fhir/sid/ndc"


class CodeConstantsURLMappingMixin:
    """A class that maps code systems to their URLs."""

    CODE_SYSTEM_MAPPING = {
        CodeConstants.CPT: CodeConstants.URL_CPT,
        CodeConstants.HCPCS: CodeConstants.URL_HCPCS,
        CodeConstants.HCPCSLEVELII: CodeConstants.URL_HCPCSLEVELII,
        CodeConstants.CVX: CodeConstants.URL_CVX,
        CodeConstants.LOINC: CodeConstants.URL_LOINC,
        CodeConstants.SNOMEDCT: CodeConstants.URL_SNOMEDCT,
        CodeConstants.FDB: CodeConstants.URL_FDB,
        CodeConstants.RXNORM: CodeConstants.URL_RXNORM,
        CodeConstants.ICD10CM: CodeConstants.URL_ICD10,
        CodeConstants.ICD10PCS: CodeConstants.URL_ICD10,
        CodeConstants.NUCC: CodeConstants.URL_NUCC,
        CodeConstants.CANVAS: CodeConstants.URL_CANVAS,
        CodeConstants.INTERNAL: CodeConstants.URL_INTERNAL,
        CodeConstants.NDC: CodeConstants.URL_NDC,
    }


class CombinedValueSet(CodeConstantsURLMappingMixin):
    """A class representing a combination of two value sets."""

    def __init__(
        self,
        value_set_1: Union[type["ValueSet"], "CombinedValueSet"],
        value_set_2: Union[type["ValueSet"], "CombinedValueSet"],
    ) -> None:
        self.value_set_1 = value_set_1
        self.value_set_2 = value_set_2

    @property
    def values(self) -> dict[str, set]:
        """A property that returns the combined values from both value sets."""
        values: dict[str, set] = defaultdict(set)

        for vs in [self.value_set_1, self.value_set_2]:
            sub_values = vs.values

            for key in sub_values:
                values[key] |= sub_values[key]

        return values

    def __or__(self, value_set: Union[type["ValueSet"], "CombinedValueSet"]) -> "CombinedValueSet":
        """Implements the `|` (or) operator to combine value sets."""
        return CombinedValueSet(self, value_set)


class ValueSystems(type):
    """A metaclass for defining a ValueSet."""

    def __or__(self, value_set: Union[type["ValueSet"], "CombinedValueSet"]) -> "CombinedValueSet":  # type: ignore[override]
        """Implements the `|` (or) operator."""
        return CombinedValueSet(cast(type["ValueSet"], self), value_set)

    def __ror__(self, value_set: Union[type["ValueSet"], "CombinedValueSet"]) -> "CombinedValueSet":  # type: ignore[override]
        """Implements the `|` (or) operator."""
        return self.__or__(value_set)


class ValueSet(CodeConstantsURLMappingMixin, metaclass=ValueSystems):
    """The Base class for a ValueSet."""

    @classproperty
    def values(cls) -> dict[str, set[str]]:
        """A property that returns a dictionary of code systems and their associated values."""
        return {
            system: getattr(cls, system)
            for system in cls.CODE_SYSTEM_MAPPING
            if hasattr(cls, system)
        }



from abc import ABC, abstractmethod

from canvas_sdk.effects import Effect
from canvas_sdk.events import EventType
from canvas_sdk.handlers import BaseHandler


class Application(BaseHandler, ABC):
    """An embeddable application that can be registered to Canvas."""

    RESPONDS_TO = [EventType.Name(EventType.APPLICATION__ON_OPEN)]

    def compute(self) -> list[Effect]:
        """Handle the application events."""
        match self.event.type:
            case EventType.APPLICATION__ON_OPEN:
                return [self.on_open()] if self.event.target.id == self.identifier else []
            case _:
                return []

    @abstractmethod
    def on_open(self) -> Effect:
        """Handle the application open event."""
        ...

    @property
    def identifier(self) -> str:
        """The application identifier."""
        return f"{self.__class__.__module__}:{self.__class__.__qualname__}"



from datetime import date, datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, ConfigDict
from pydantic_core import InitErrorDetails, PydanticCustomError, ValidationError


class Model(BaseModel):
    """A base model that includes validation methods."""

    class Meta:
        pass

    model_config = ConfigDict(
        strict=True,
        revalidate_instances="always",
        validate_assignment=True,
        json_encoders={
            date: lambda v: v.isoformat(),
            datetime: lambda v: v.isoformat(),
            Enum: lambda v: v.value,
        },
    )

    def _get_effect_method_required_fields(self, method: Any) -> tuple:
        return getattr(self.Meta, f"{method}_required_fields", ())

    def _create_error_detail(self, type: str, message: str, value: Any) -> InitErrorDetails:
        return InitErrorDetails({"type": PydanticCustomError(type, message), "input": value})

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        required_fields = self._get_effect_method_required_fields(method)
        class_name = self.__repr_name__()  # type: ignore[misc]

        class_name_article = "an" if class_name.startswith(("A", "E", "I", "O", "U")) else "a"

        error_details = []
        for field in required_fields:
            fields = field.split("|")
            if not all(getattr(self, f) is None for f in fields):
                continue
            field_description = " or ".join([f"'{f}'" for f in fields])
            message = f"Field {field_description} is required to {method.replace('_', ' ')} {class_name_article} {class_name}"
            error = self._create_error_detail("missing", message, None)
            error_details.append(error)

        return error_details

    def _validate_before_effect(self, method: str) -> None:
        self.model_validate(self)
        if error_details := self._get_error_details(method):
            raise ValidationError.from_exception_data(self.__class__.__name__, error_details)



from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.commands.goal import GoalCommand


class CloseGoalCommand(BaseCommand):
    """A class for managing a CloseGoal command within a specific note."""

    class Meta:
        key = "closeGoal"
        commit_required_fields = ("goal_id",)

    goal_id: int | None = None
    achievement_status: GoalCommand.AchievementStatus | None = None
    progress: str | None = None

    @property
    def values(self) -> dict:
        """The CloseGoal command's field values."""
        return {
            "goal_id": self.goal_id,
            "achievement_status": (
                self.achievement_status.value if self.achievement_status else None
            ),
            "progress": self.progress,
        }
